<table>
<thead>
</thead>
<tbody>
<tr><td>ID</td><td>{{ task.task_id }}</td></tr>
<tr><td>Info</td><td>{{ task.info }}</td></tr>
<tr><td>Step</td><td>{{ task_class }}</td></tr>
<tr><td>Source</td><td colspan="2">
{% with elements=source_elements %}
{% include "assetset_view.html"%}
{% endwith %}
</td>
</tr>
<tr><td>Target</td>
<td colspan="2">
{% with elements = target_elements %}
{% include "assetset_view.html" %}
{% endwith %}
</td>
</tr>
<tr><td>Parameters</td><td>{{ parameters }}</td></tr>
</tbody>
</table>
